{
* TGrep 1.0 beta
* 
* Author:
*   Gabriele Carioli 
*   Viale Antonio Gramsci, 102
*   47100 Forl (FO) - ITALY
*   E-mail: bilo@lingue.unibo.it
* 
* Copyright  1997-98 by Gabriele Carioli <bilo@lingue.unibo.it>
* 
* This software is provided as it is, without any kind of warranty
* given. Use it at your own risk.
* 
* You may use this software in any kind of development, including
* comercial, redistribute, and modify it freely, under the 
* following restrictions :
* 
* 1. The origin of this software may not be mispresented, you must
*    not claim that you wrote the original software. If you use
*    this software in any kind of product, it would be appreciated
*    that there in a information box, or in the documentation would
*    be an acknowledgmnent like this
*           Parts Copyright  1998 by Gabriele Carioli
* 
* 2. You may not have any income from distributing this source
*    to other developers. When you use this product in a comercial
*    package, the source may not be charged seperatly.
* 
* 3. This notice should also follow the package if you use this
*    component in a commercial product.
}

unit Grep;

interface

uses
  SysUtils, Classes, DsgnIntF;


const
  // A collection of sample regular expressions
  cTCPAddr  = '[_a-zA-Z0-9\-]+(\.[_a-zA-Z0-9\-]+)+';
  cMailAddr = '([_a-zA-Z0-9\-\.]+)@(' + cTCPAddr + ')';
  cUnixPath = '(/[ _a-zA-Z0-9\-\\\.]+)+';
  cURL      = '([Ff][Tt][Pp]|[Hh][Tt][Tt][Pp])://(' + cTCPAddr + ')(' + cUnixPath + ')';

  NSUBEXP = 10; // Do NOT change!

type

  ERegErr = class(Exception);

  TSubExpRange = 0..NSUBEXP - 1;

  TRegExp = string;

  TRegExpC = record
    startp:   array[TSubExpRange] of pchar;
    endp:     array[TSubExpRange] of pchar;
    regstart: char;                         // Internal use
    reganch:  char;                         // Internal use
    regmust:  pchar;                        // Internal use
    regmlen:  integer;                      // Internal use
    progr:    array[0..0] of char;          // Internal use (single element)
  end;
  PTRegExpC = ^TRegExpC;

  TRegErrProc = procedure (buf: pchar); stdcall;

  TSubRegExp = class
    FStartP: integer;
    FEndP: integer;
    FText: string;
  protected
    procedure SetAll(s, e: integer; t: string);
    constructor Create;
  published
    property StartP: integer read FStartP;
    property EndP: integer read FEndP;
    property Text: string read FText;
  end;

  TGrep = class(TComponent)
  private
    FProg: PTRegExpC;
    FRegExp: TRegExp;
    FText: string;
    FPText: pchar;
    FPTextPos: pchar;
    FFound: boolean;
    FSubExpRetrieved: boolean;
    FSubExp: array[TSubExpRange] of TSubRegExp;
    FOnGrep: TNotifyEvent;
    procedure SetRegExp(s: string);
    procedure SetText(s: string);
    function GetSubExp(index: TSubExpRange): TSubRegExp;
    procedure Reset;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Grep: boolean;
    function Subst(s: string): string;
    procedure Split(max: integer; elems: TStrings);

  published
    property RegExp: string read FRegExp write SetRegExp;
    property Text: string read FText write SetText;
    property SubExp[index: TSubExpRange]: TSubRegExp read GetSubExp; default;
    property OnGrep: TNotifyEvent read FOnGrep write FOnGrep;
    property Found: boolean read FFound;
  end;

  TRegExpProperty = class(TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;

  procedure Register;

implementation

uses
  Forms, Controls, RegExpForm;


function RegComp(exp: pchar): PTRegExpC; stdcall; external 'regexp.dll' name 'regcomp';
procedure RegDestroy(prog: PTRegExpC); stdcall; external 'regexp.dll' name 'regdestroy';
function RegExec(prog: PTRegExpC; str: pchar): integer; stdcall; external 'regexp.dll' name 'regexec';
procedure RegisterRegErrProc(RegErrProc: TRegErrProc); stdcall; external 'regexp.dll' name 'registerregerrproc';

procedure ErrProc(p: pchar); stdcall; forward;

constructor TSubRegExp.Create;
begin
  FStartP := 0;
  FEndP := 0;
  FText := '';
end;

procedure TSubRegExp.SetAll(s, e: integer; t: string);
begin
  FStartP := s;
  FEndP := e;
  FText := t;
end;

destructor TGrep.Destroy;
begin
  if FProg <> nil then RegDestroy(FProg);
  inherited Destroy;
end;

constructor TGrep.Create(AOwner: TComponent);
var
  i: integer;
begin
  inherited Create(AOwner);
  FProg := nil;
  FText := '';
  FPText := nil;
  FPTextPos := nil;
  FFound := false;
  FSubExpRetrieved := false;
  for i := 0 to NSUBEXP - 1 do FSubExp[i] := TSubRegExp.Create;
end;

procedure TGrep.Reset;
begin
  FPTextPos := FPText;
  FFound := false;
  FSubExpRetrieved := false;
end;

procedure TGrep.Split(max: integer; elems: TStrings);
var
  p: integer;
  n, l: integer;
  sub: string;
begin
  if not Assigned(elems) then
    raise ERegErr.Create('ERegErr: undefined TStrings')
  else begin
    Reset;
    l := length(Text);
    if max = 0 then n := -1 else n := 0;
    p := 1;
    while (p <= l) and (n < max) do begin
      if max > 0 then n := n + 1;
      if Grep and (n < max) then
        begin
          sub := copy(Text, p, SubExp[0].StartP - p);
          p := SubExp[0].EndP;
        end
      else
        begin
          sub := copy(Text, p, l - p + 1);
          p := l + 1;
        end;
      elems.add(sub);
    end;
    Reset;
  end;

end;

function TGrep.GetSubExp(index: TSubExpRange): TSubRegExp;
var
  i: TSubExpRange;
begin
  if FSubExpRetrieved then
    Result := FSubExp[index]
  else
    begin
      if FFound then
        for i := 0 to NSUBEXP - 1 do
          if FProg^.StartP[i] <> nil then
            FSubExp[i].SetAll(
              FProg^.StartP[i] - FPText + 1,
              FProg^.EndP[i] - FPText + 1,
              copy(Text, FProg^.StartP[i] - FPText + 1, FProg^.EndP[i] - FProg^.StartP[i])
            )
          else
            FSubExp[i].SetAll(0, 0, '')
      else
        for i := 0 to NSUBEXP - 1 do
          FSubExp[i].SetAll(0, 0, '');

      FSubExpRetrieved := true;
      Result := FSubExp[index]
    end;
end;

function TGrep.Subst(s: string): string;
var
  i: integer;
  r: string;
begin
  r := '';
  if FFound and (length(s) > 0) then begin
    i := 1;
    while i <= length(s) do begin
      case s[i] of
        '&': r := r + SubExp[0].Text;
        '\': if i = length(s) then
               raise ERegErr.Create('ERegErr: damaged match string')
             else
               begin
                 i := i + 1;
                 if s[i] in ['0'..'9']
                   then r := r + SubExp[ord(s[i]) - ord('0')].Text
                   else r := r + s[i];
               end;
        else r := r + s[i];
      end;
      i := i + 1;
    end; {while}
  end;
  Result := r;
end;

function TGrep.Grep: boolean;
begin
  if (FPTextPos <> nil) and (FPTextPos^ <> #0) and (FProg <> nil) then
    begin
      Result := RegExec(FProg, FPTextPos) <> 0;
      if Result
        then
          if FPTextPos = FProg^.EndP[0]
            then FPTextPos := FProg^.EndP[0] + 1
            else FPTextPos := FProg^.EndP[0]
        else FPTextPos := nil;
    end
  else
    Result := false;
  FFound := Result;
  FSubExpRetrieved := False;

  if Assigned(FOnGrep) then FOnGrep(Self);

end;

procedure TGrep.SetRegExp(s:string);
begin
  if FProg <> nil then RegDestroy(FProg);
//if s <> '' then begin
  FRegExp := s;
  FProg := RegComp(pchar(FRegExp));
//end;
  Reset;
end;

procedure TGrep.SetText(s: string);
begin
  if Length(s) > 0 then
    begin
      FText := s;
      FPText := pchar(FText);
    end
  else
    begin
      FText := '';
      FPText := nil;
     end;
  Reset;
end;

procedure ErrProc(p: pchar); stdcall;
begin
  raise ERegErr.Create(format('ERegErr: %s', [string(p)]));
end;

function TRegExpProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paReadOnly];
end;

procedure TRegExpProperty.Edit;
var
  g: TGrep;
  peForm: TRegExpEditor;
begin
  g := GetComponent(0) as TGrep;
  try
    peForm := TRegExpEditor.Create(Application);
    peForm.txtRegExp.Text := g.RegExp;
    if peForm.ShowModal = mrOk then g.RegExp := peForm.txtRegExp.Text;
  finally
    peForm.Free;
  end;
end;

procedure Register;
begin
  RegisterComponents('Additional...', [TGrep]);
  RegisterPropertyEditor(TypeInfo(TRegExp), TGrep, 'RegExp', TRegExpProperty);
end;

initialization
  RegisterRegErrProc(ErrProc);
end.

